#include <iostream>
#include <cstdint>
#include <vector>
#include <array>
#include <volk.h>

//########################################################

VkInstance vkInstance = nullptr;
VkPhysicalDevice vkPhysicalDevice = nullptr;
VkDevice vkDevice = nullptr;

uint32_t vkQueueFamilyIndex = 0;
VkQueue vkQueue = nullptr;

VkVideoEncodeH264ProfileEXT vkEncodeProfile;
VkVideoProfileKHR vkVideoProfile;

VkVideoEncodeH264CapabilitiesEXT vkEncodeCapabilities;
VkVideoCapabilitiesKHR vkVideoCapabilities;

VkExtent2D vkImageExtend;
VkExtent2D vkBlockCount;
uint32_t vkBufferSize = 0;

VkVideoSessionKHR vkVideoSession = nullptr;
std::vector<VkDeviceMemory> vkVideoSessionMemory;
VkVideoSessionParametersKHR vkVideoParameters = nullptr;

VkImage vkInputImage = nullptr;
VkImageView vkInputView = nullptr;
VkDeviceMemory vkInputMemory = nullptr;

VkImage vkSlotImage = nullptr;
VkImageView vkSlotView = nullptr;
VkDeviceMemory vkSlotMemory = nullptr;

VkBuffer vkOutputBuffer = nullptr;
VkDeviceMemory vkOutputMemory = nullptr;

VkCommandPool vkCommandPool = nullptr;
VkCommandBuffer vkCommandBuffer = nullptr;

VkFence vkFence = nullptr;

//########################################################

bool CheckInstanceLayerSupport(const std::vector<const char*>& layerNameList)
{
	uint32_t layerCount = 0;
	VkLayerProperties* layerList = nullptr;

	vkEnumerateInstanceLayerProperties(&layerCount, nullptr);

	layerList = new VkLayerProperties[layerCount];

	vkEnumerateInstanceLayerProperties(&layerCount, layerList);

	for(uint32_t index = 0; index < layerNameList.size(); index++)
	{
		bool found = false;

		for(uint32_t layerIndex = 0; layerIndex < layerCount; layerIndex++)
		{
			if(strcmp(layerNameList[index], layerList[layerIndex].layerName) == 0)
			{
				found = true;

				break;
			}
		}

		if(!found)
		{
			std::cout << layerNameList[index] << " not supported" << std::endl;

			delete[] layerList;

			return false;
		}
	}

	delete[] layerList;

	return true;
}

bool CheckDeviceLayerSupport(VkPhysicalDevice physicalDevice, const std::vector<const char*>& layerNameList)
{
	uint32_t layerCount = 0;
	VkLayerProperties* layerList = nullptr;

	vkEnumerateDeviceLayerProperties(physicalDevice, &layerCount, nullptr);
	
	layerList = new VkLayerProperties[layerCount];

	vkEnumerateDeviceLayerProperties(physicalDevice, &layerCount, layerList);

	for(uint32_t index = 0; index < layerNameList.size(); index++)
	{
		bool found = false;

		for(uint32_t layerIndex = 0; layerIndex < layerCount; layerIndex++)
		{
			if(strcmp(layerNameList[index], layerList[layerIndex].layerName) == 0)
			{
				found = true;

				break;
			}
		}

		if(!found)
		{
			std::cout << layerNameList[index] << " not supported" << std::endl;

			delete[] layerList;

			return false;
		}
	}

	delete[] layerList;

	return true;
}

bool CheckInstanceExtensionSupport(const std::vector<const char*>& extensionNameList)
{
	uint32_t extensionCount = 0;
	VkExtensionProperties* extensionList = nullptr;

	vkEnumerateInstanceExtensionProperties(nullptr, &extensionCount, nullptr);

	extensionList = new VkExtensionProperties[extensionCount];

	vkEnumerateInstanceExtensionProperties(nullptr, &extensionCount, extensionList);

	for(uint32_t index = 0; index < extensionNameList.size(); index++)
	{
		bool found = false;

		for(uint32_t extensionIndex = 0; extensionIndex < extensionCount; extensionIndex++)
		{
			if(strcmp(extensionNameList[index], extensionList[extensionIndex].extensionName) == 0)
			{
				found = true;

				break;
			}
		}

		if(!found)
		{
			std::cout << extensionNameList[index] << " not supported" << std::endl;

			delete[] extensionList;

			return false;
		}
	}

	delete[] extensionList;

	return true;
}

bool CheckDeviceExtensionSupport(VkPhysicalDevice physicalDevice, const std::vector<const char*>& extensionNameList)
{
	uint32_t extensionCount = 0;
	VkExtensionProperties* extensionList = nullptr;

	vkEnumerateDeviceExtensionProperties(physicalDevice, nullptr, &extensionCount, nullptr);

	extensionList = new VkExtensionProperties[extensionCount];

	vkEnumerateDeviceExtensionProperties(physicalDevice, nullptr, &extensionCount, extensionList);

	for(uint32_t index = 0; index < extensionNameList.size(); index++)
	{
		bool found = false;

		for(uint32_t extensionIndex = 0; extensionIndex < extensionCount; extensionIndex++)
		{
			if(strcmp(extensionNameList[index], extensionList[extensionIndex].extensionName) == 0)
			{
				found = true;

				break;
			}
		}

		if(!found)
		{
			std::cout << extensionNameList[index] << "not supported" << std::endl;

			delete[] extensionList;

			return false;
		}
	}

	delete[] extensionList;

	return true;
}

uint32_t SelectQueueFamiliy(VkPhysicalDevice physicalDevice)
{
	uint32_t queueFamilyCount;
	VkQueueFamilyProperties* queueFamilyList;

	vkGetPhysicalDeviceQueueFamilyProperties(physicalDevice, &queueFamilyCount, nullptr);

	queueFamilyList = new VkQueueFamilyProperties[queueFamilyCount];

	vkGetPhysicalDeviceQueueFamilyProperties(physicalDevice, &queueFamilyCount, queueFamilyList);

	uint32_t queueFamiliyIndex = 0xFFFFFFFF;

	for(uint32_t index = 0; index < queueFamilyCount; index++)
	{
		if((queueFamilyList[index].queueFlags & VK_QUEUE_VIDEO_ENCODE_BIT_KHR) != 0 && queueFamilyList[index].queueCount > 0)
		{
			queueFamiliyIndex = index;

			break;
		}
	}

	return queueFamiliyIndex;
}

VkPhysicalDevice SelectPhysicalDevice()
{
	uint32_t physicalDeviceCount = 0;
	VkPhysicalDevice* physicalDeviceList = nullptr;

	vkEnumeratePhysicalDevices(vkInstance, &physicalDeviceCount, nullptr);

	physicalDeviceList = new VkPhysicalDevice[physicalDeviceCount];

	vkEnumeratePhysicalDevices(vkInstance, &physicalDeviceCount, physicalDeviceList);

	VkPhysicalDevice physicalDevice = nullptr;

	for(uint32_t index = 0; index < physicalDeviceCount; index++)
	{
		VkPhysicalDeviceProperties physicalDeviceProperties;
		VkPhysicalDeviceFeatures physicalDeviceFeatures;

		vkGetPhysicalDeviceProperties(physicalDeviceList[index], &physicalDeviceProperties);
		vkGetPhysicalDeviceFeatures(physicalDeviceList[index], &physicalDeviceFeatures);

		if(physicalDeviceProperties.deviceType == VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU)
		{
			if(SelectQueueFamiliy(physicalDeviceList[index]) != 0xFFFFFFFF)
			{
				physicalDevice = physicalDeviceList[index];

				break;
			}
		}
	}

	delete[] physicalDeviceList;

	return physicalDevice;
}

bool SelectMemory(uint32_t filterMask, VkMemoryPropertyFlags flags, uint32_t& memoryTypeIndex)
{
	VkPhysicalDeviceMemoryProperties memoryProperties;

	vkGetPhysicalDeviceMemoryProperties(vkPhysicalDevice, &memoryProperties);

	for(uint32_t index = 0; index < memoryProperties.memoryTypeCount; index++)
	{
		if((filterMask & (1 << index)) != 0 && (memoryProperties.memoryTypes[index].propertyFlags & flags) == flags)
		{
			memoryTypeIndex = index;

			return true;
		}
	}

	return false;
}

bool CheckEncodeSupport(VkPhysicalDevice physicalDevice, uint32_t queueFamilyIndex)
{
	uint32_t property_count = 0;
	vkGetPhysicalDeviceQueueFamilyProperties2(physicalDevice, &property_count, nullptr);

	std::vector<VkVideoQueueFamilyProperties2KHR> video_property_list;
	video_property_list.resize(property_count);
	std::vector<VkQueueFamilyProperties2> property_list;
	property_list.resize(property_count);

	for(uint32_t index = 0; index < property_count; index++)
	{
		video_property_list[index].sType = VK_STRUCTURE_TYPE_VIDEO_QUEUE_FAMILY_PROPERTIES_2_KHR;
		video_property_list[index].pNext = nullptr;
		property_list[index].sType = VK_STRUCTURE_TYPE_QUEUE_FAMILY_PROPERTIES_2;
		property_list[index].pNext = &video_property_list[index];
	}

	vkGetPhysicalDeviceQueueFamilyProperties2(physicalDevice, &property_count, property_list.data());

	if((video_property_list[queueFamilyIndex].videoCodecOperations & VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_EXT) != 0)
	{
		return true;
	}

	return false;
}

bool CheckFormatSupport(VkPhysicalDevice physicalDevice, VkImageUsageFlags usage, VkFormat format)
{
	VkFormatProperties2 format_properties;
	format_properties.sType = VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_2;
	format_properties.pNext = &vkVideoProfile;

	vkGetPhysicalDeviceFormatProperties2(physicalDevice, format, &format_properties);

	/*if((format_properties.formatProperties.optimalTilingFeatures & VK_FORMAT_FEATURE_VIDEO_ENCODE_INPUT_BIT_KHR) == 0)
	{
		return false;
	}

	if((format_properties.formatProperties.optimalTilingFeatures & VK_FORMAT_FEATURE_VIDEO_ENCODE_DPB_BIT_KHR) == 0)
	{
		return false;
	}*/

	VkPhysicalDeviceImageFormatInfo2 image_format_info;
	image_format_info.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_FORMAT_INFO_2;
	image_format_info.pNext = nullptr;
	image_format_info.format = format;
	image_format_info.type = VK_IMAGE_TYPE_2D;
	image_format_info.tiling = VK_IMAGE_TILING_OPTIMAL;
	image_format_info.usage = usage;
	image_format_info.flags = VK_IMAGE_CREATE_ALIAS_BIT | VK_IMAGE_CREATE_DISJOINT_BIT;

	VkImageFormatProperties2 image_format_properties;
	image_format_properties.sType = VK_STRUCTURE_TYPE_IMAGE_FORMAT_PROPERTIES_2;
	image_format_properties.pNext = nullptr;

	if(vkGetPhysicalDeviceImageFormatProperties2(physicalDevice, &image_format_info, &image_format_properties) != VK_SUCCESS)
	{
		return false;
	}

	if(image_format_properties.imageFormatProperties.maxExtent.width == 0 || image_format_properties.imageFormatProperties.maxExtent.height == 0)
	{
		return false;
	}

	VkVideoProfilesKHR video_profiles;
	video_profiles.sType = VK_STRUCTURE_TYPE_VIDEO_PROFILES_KHR;
	video_profiles.pNext = nullptr;
	video_profiles.profileCount = 1;
	video_profiles.pProfiles = &vkVideoProfile;

	VkPhysicalDeviceVideoFormatInfoKHR video_format_info;
	video_format_info.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_FORMAT_INFO_KHR;
	video_format_info.pNext = nullptr;
	video_format_info.imageUsage = usage;
	video_format_info.pVideoProfiles = &video_profiles;

	uint32_t format_count = 0;

	if(vkGetPhysicalDeviceVideoFormatPropertiesKHR(physicalDevice, &video_format_info, &format_count, nullptr) != VK_SUCCESS)
	{
		return false;
	}

	std::vector<VkVideoFormatPropertiesKHR> format_list;
	format_list.resize(format_count);

	for(VkVideoFormatPropertiesKHR& format_properties : format_list)
	{
		format_properties.sType = VK_STRUCTURE_TYPE_VIDEO_FORMAT_PROPERTIES_KHR;
		format_properties.pNext = nullptr;
	}

	if(vkGetPhysicalDeviceVideoFormatPropertiesKHR(physicalDevice, &video_format_info, &format_count, format_list.data()) != VK_SUCCESS)
	{
		return false;
	}

	for(const VkVideoFormatPropertiesKHR& format_properties : format_list)
	{
		if(format_properties.format == format)
		{
			return true;
		}
	}

	return false;
}

//########################################################

bool CreateInstance()
{
	const std::vector<const char*> instanceLayerList =
	{
		"VK_LAYER_KHRONOS_validation"
	};

	if(!CheckInstanceLayerSupport(instanceLayerList))
	{
		return false;
	}

	const std::vector<const char*> instanceExtensionList =
	{
	};

	if(!CheckInstanceExtensionSupport(instanceExtensionList))
	{
		return false;
	}

	VkApplicationInfo applicationInfo;
	applicationInfo.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO;
	applicationInfo.pNext = nullptr;
	applicationInfo.pApplicationName = "Vulkan Video";
	applicationInfo.applicationVersion = VK_MAKE_VERSION(1, 0, 0);
	applicationInfo.pEngineName = "No Engine";
	applicationInfo.engineVersion = VK_MAKE_VERSION(1, 0, 0);
	applicationInfo.apiVersion = VK_API_VERSION_1_1;

	VkInstanceCreateInfo instanceCreateInfo;
	instanceCreateInfo.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
	instanceCreateInfo.pNext = nullptr;
	instanceCreateInfo.flags = 0;
	instanceCreateInfo.pApplicationInfo = &applicationInfo;
	instanceCreateInfo.enabledLayerCount = (uint32_t)instanceLayerList.size();
	instanceCreateInfo.ppEnabledLayerNames = instanceLayerList.data();
	instanceCreateInfo.enabledExtensionCount = (uint32_t)instanceExtensionList.size();
	instanceCreateInfo.ppEnabledExtensionNames = instanceExtensionList.data();

	VkResult result = vkCreateInstance(&instanceCreateInfo, nullptr, &vkInstance);

	if(result != VK_SUCCESS)
	{
		std::cout << "Can't create Instance" << std::endl;

		return false;
	}

	return true;
}

bool CreateDevice()
{
	vkPhysicalDevice = SelectPhysicalDevice();

	if(vkPhysicalDevice == nullptr)
	{
		return false;
	}

	uint32_t queueFamiliyIndex = SelectQueueFamiliy(vkPhysicalDevice);

	if(queueFamiliyIndex == 0xFFFFFFFF)
	{
		return false;
	}

	float queuePriority = 1.0f;

	VkDeviceQueueCreateInfo deviceQueueCreateInfo;
	deviceQueueCreateInfo.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
	deviceQueueCreateInfo.pNext = nullptr;
	deviceQueueCreateInfo.flags = 0;
	deviceQueueCreateInfo.queueFamilyIndex = queueFamiliyIndex;
	deviceQueueCreateInfo.queueCount = 1;
	deviceQueueCreateInfo.pQueuePriorities = &queuePriority;

	VkPhysicalDeviceFeatures deviceFeatures;
	memset(&deviceFeatures, 0, sizeof(deviceFeatures));

	const std::vector<const char*> deviceLayerList =
	{
		"VK_LAYER_KHRONOS_validation"
	};

	if(!CheckDeviceLayerSupport(vkPhysicalDevice, deviceLayerList))
	{
		return false;
	}

	const std::vector<const char*> deviceExtensionList =
	{
		VK_KHR_VIDEO_QUEUE_EXTENSION_NAME,
		VK_KHR_VIDEO_ENCODE_QUEUE_EXTENSION_NAME,
		VK_KHR_SYNCHRONIZATION_2_EXTENSION_NAME,
		VK_EXT_VIDEO_ENCODE_H264_EXTENSION_NAME
	};

	if(!CheckDeviceExtensionSupport(vkPhysicalDevice, deviceExtensionList))
	{
		return false;
	}

	VkDeviceCreateInfo deviceCreateInfo;
	deviceCreateInfo.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
	deviceCreateInfo.pNext = nullptr;
	deviceCreateInfo.flags = 0;
	deviceCreateInfo.queueCreateInfoCount = 1;
	deviceCreateInfo.pQueueCreateInfos = &deviceQueueCreateInfo;
	deviceCreateInfo.enabledLayerCount = (uint32_t)deviceLayerList.size();
	deviceCreateInfo.ppEnabledLayerNames = deviceLayerList.data();
	deviceCreateInfo.enabledExtensionCount = (uint32_t)deviceExtensionList.size();
	deviceCreateInfo.ppEnabledExtensionNames = deviceExtensionList.data();
	deviceCreateInfo.pEnabledFeatures = &deviceFeatures;

	VkResult result = vkCreateDevice(vkPhysicalDevice, &deviceCreateInfo, nullptr, &vkDevice);

	if(result != VK_SUCCESS)
	{
		std::cout << "Can't create Device" << std::endl;

		return false;
	}

	vkGetDeviceQueue(vkDevice, queueFamiliyIndex, 0, &vkQueue);

	vkQueueFamilyIndex = queueFamiliyIndex;

	return true;
}

bool CheckProfiles()
{
	vkEncodeProfile.sType = VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_PROFILE_EXT;
	vkEncodeProfile.pNext = nullptr;
	vkEncodeProfile.stdProfileIdc = STD_VIDEO_H264_PROFILE_IDC_BASELINE;

	vkVideoProfile.sType = VK_STRUCTURE_TYPE_VIDEO_PROFILE_KHR;
	vkVideoProfile.pNext = &vkEncodeProfile;
	vkVideoProfile.videoCodecOperation = VK_VIDEO_CODEC_OPERATION_ENCODE_H264_BIT_EXT;
	vkVideoProfile.chromaSubsampling = VK_VIDEO_CHROMA_SUBSAMPLING_444_BIT_KHR;
	vkVideoProfile.lumaBitDepth = VK_VIDEO_COMPONENT_BIT_DEPTH_8_BIT_KHR;
	vkVideoProfile.chromaBitDepth = VK_VIDEO_COMPONENT_BIT_DEPTH_8_BIT_KHR;

	vkEncodeCapabilities.sType = VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_CAPABILITIES_EXT;
	vkEncodeCapabilities.pNext = nullptr;

	vkVideoCapabilities.sType = VK_STRUCTURE_TYPE_VIDEO_CAPABILITIES_KHR;
	vkVideoCapabilities.pNext = &vkEncodeCapabilities;

	if(vkGetPhysicalDeviceVideoCapabilitiesKHR(vkPhysicalDevice, &vkVideoProfile, &vkVideoCapabilities) != VK_SUCCESS)
	{
		std::cout << "Can't Create Profile" << std::endl;

		return false;
	}

	if(!CheckEncodeSupport(vkPhysicalDevice, vkQueueFamilyIndex))
	{
		return false;
	}

	if(!CheckFormatSupport(vkPhysicalDevice, VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR, VK_FORMAT_G8_B8R8_2PLANE_444_UNORM_EXT))
	{
		return false;
	}

	if(!CheckFormatSupport(vkPhysicalDevice, VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR, VK_FORMAT_G8_B8R8_2PLANE_444_UNORM_EXT))
	{
		return false;
	}

	return true;
}

void ComputeDimensions()
{
	vkImageExtend = vkVideoCapabilities.minExtent;
	vkBlockCount.width = vkImageExtend.width / 16;
	vkBlockCount.height = vkImageExtend.height / 16;

	vkBufferSize = (uint32_t)vkVideoCapabilities.minBitstreamBufferSizeAlignment * 100000;
}

bool CreateSession()
{
	VkVideoEncodeH264SessionCreateInfoEXT encode_create_info;
	encode_create_info.sType = VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_CREATE_INFO_EXT;
	encode_create_info.pNext = nullptr;
	encode_create_info.flags = VK_VIDEO_ENCODE_H264_CREATE_RESERVED_0_BIT_EXT;
	encode_create_info.maxPictureSizeInMbs = vkBlockCount;
	encode_create_info.pStdExtensionVersion = &vkEncodeCapabilities.stdExtensionVersion;

	VkVideoSessionCreateInfoKHR video_create_info;
	video_create_info.sType = VK_STRUCTURE_TYPE_VIDEO_SESSION_CREATE_INFO_KHR;
	video_create_info.pNext = &encode_create_info;
	video_create_info.queueFamilyIndex = vkQueueFamilyIndex;
	video_create_info.flags = VK_VIDEO_SESSION_CREATE_DEFAULT_KHR;
	video_create_info.pVideoProfile = &vkVideoProfile;
	video_create_info.pictureFormat = VK_FORMAT_G8_B8R8_2PLANE_444_UNORM_EXT;
	video_create_info.maxCodedExtent = vkImageExtend;
	video_create_info.referencePicturesFormat = VK_FORMAT_G8_B8R8_2PLANE_444_UNORM_EXT;
	video_create_info.maxReferencePicturesSlotsCount = 1;
	video_create_info.maxReferencePicturesActiveCount = 1;

	if(vkCreateVideoSessionKHR(vkDevice, &video_create_info, nullptr, &vkVideoSession) != VK_SUCCESS)
	{
		return false;
	}

	return true;
}

bool BindSessionMemory()
{
	uint32_t property_count = 0;
	if(vkGetVideoSessionMemoryRequirementsKHR(vkDevice, vkVideoSession, &property_count, nullptr) != VK_SUCCESS)
	{
		return false;
	}

	std::vector<VkMemoryRequirements2> requirement_list;
	requirement_list.resize(property_count);
	std::vector<VkVideoGetMemoryPropertiesKHR> property_list;
	property_list.resize(property_count);

	for(uint32_t index = 0; index < property_count; index++)
	{
		requirement_list[index].sType = VK_STRUCTURE_TYPE_MEMORY_REQUIREMENTS_2;
		requirement_list[index].pNext = nullptr;
		property_list[index].sType = VK_STRUCTURE_TYPE_VIDEO_GET_MEMORY_PROPERTIES_KHR;
		property_list[index].pNext = nullptr;
		property_list[index].pMemoryRequirements = &requirement_list[index];
	}

	if(vkGetVideoSessionMemoryRequirementsKHR(vkDevice, vkVideoSession, &property_count, property_list.data()) != VK_SUCCESS)
	{
		return false;
	}

	std::vector<VkVideoBindMemoryKHR> bind_list;

	for(const VkVideoGetMemoryPropertiesKHR& property : property_list)
	{
		VkMemoryRequirements requirements = property.pMemoryRequirements->memoryRequirements;

		uint32_t memoryTypeIndex = 0;

		if(!SelectMemory(requirements.memoryTypeBits, 0, memoryTypeIndex))
		{
			return false;
		}

		VkMemoryAllocateInfo memoryAllocateInfo;
		memoryAllocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		memoryAllocateInfo.pNext = nullptr;
		memoryAllocateInfo.allocationSize = requirements.size;
		memoryAllocateInfo.memoryTypeIndex = memoryTypeIndex;

		VkDeviceMemory vkDeviceMemory = nullptr;

		if(vkAllocateMemory(vkDevice, &memoryAllocateInfo, nullptr, &vkDeviceMemory) != VK_SUCCESS)
		{
			return false;
		}

		vkVideoSessionMemory.push_back(vkDeviceMemory);

		VkVideoBindMemoryKHR bind;
		bind.sType = VK_STRUCTURE_TYPE_VIDEO_BIND_MEMORY_KHR;
		bind.pNext = nullptr;
		bind.memoryBindIndex = property.memoryBindIndex;
		bind.memory = vkDeviceMemory;
		bind.memoryOffset = 0;
		bind.memorySize = requirements.size;

		bind_list.push_back(bind);
	}

	if(vkBindVideoSessionMemoryKHR(vkDevice, vkVideoSession, (uint32_t)bind_list.size(), bind_list.data()) != VK_SUCCESS)
	{
		return false;
	}

	return true;
}

bool CreateParameters()
{
    //H.264 Specification Section 7.4.2.1.1: 
    //https://www.itu.int/rec/dologin_pub.asp?lang=e&id=T-REC-H.264-202108-I!!PDF-E&type=items
    StdVideoH264SpsFlags sps_flags;
    sps_flags.constraint_set0_flag = 0;                                            //No constraints
    sps_flags.constraint_set1_flag = 0;                                            //No constraints
    sps_flags.constraint_set2_flag = 0;                                            //No constraints
    sps_flags.constraint_set3_flag = 0;                                            //No constraints
    sps_flags.constraint_set4_flag = 0;                                            //No constraints
    sps_flags.constraint_set5_flag = 0;                                            //No constraints
    sps_flags.direct_8x8_inference_flag = 0;                                       //Only for b-frames set it to zero
    sps_flags.mb_adaptive_frame_field_flag = 0;                                    //Set it to zero                                        
    sps_flags.frame_mbs_only_flag = 1;                                             //Only frames no fields
    sps_flags.delta_pic_order_always_zero_flag = 0;                                //Set it to zero                                        
    sps_flags.separate_colour_plane_flag = 0;                                      //Combined encoding see separate_colour_plane_flag
    sps_flags.gaps_in_frame_num_value_allowed_flag = 0;                            //No gaps
    sps_flags.qpprime_y_zero_transform_bypass_flag = 0;                            //Set it to zero                                        
    sps_flags.frame_cropping_flag = 0;                                             //Dont use Crop
    sps_flags.seq_scaling_matrix_present_flag = 0;                                 //Dont send scaling matrix
    sps_flags.vui_parameters_present_flag = 0;                                     //Dont send VUI

    StdVideoH264SequenceParameterSet sps_parameter;
    sps_parameter.profile_idc = STD_VIDEO_H264_PROFILE_IDC_BASELINE;               //Baseline profile
    sps_parameter.level_idc = STD_VIDEO_H264_LEVEL_1_0;                            //Level 1.0
    sps_parameter.seq_parameter_set_id = 0;                                        //Id of the sequence parameter set
    sps_parameter.chroma_format_idc = STD_VIDEO_H264_CHROMA_FORMAT_IDC_444;        //Chroma format 444 meaning same image size for chroma and luma
    sps_parameter.bit_depth_luma_minus8 = 0;                                       //8 bit luma
    sps_parameter.bit_depth_chroma_minus8 = 0;                                     //8 bit chroma
    sps_parameter.log2_max_frame_num_minus4 = 0;                                   //2^4 frame numbers
    sps_parameter.pic_order_cnt_type = STD_VIDEO_H264_POC_TYPE_0;                  //Picture order count type 0
    sps_parameter.log2_max_pic_order_cnt_lsb_minus4 = 0;                           //Set to zero
    sps_parameter.offset_for_non_ref_pic = 0;                                      //Set to zero
    sps_parameter.offset_for_top_to_bottom_field = 0;                              //Set to zero
    sps_parameter.num_ref_frames_in_pic_order_cnt_cycle = 0;                       //Set to zero
    sps_parameter.max_num_ref_frames = 1;                                          //1 Reference Frame
    sps_parameter.pic_width_in_mbs_minus1 = vkBlockCount.width - 1;				   //Image width in macroblocks
    sps_parameter.pic_height_in_map_units_minus1 = vkBlockCount.height - 1;		   //Image height in macroblocks
    sps_parameter.frame_crop_left_offset = 0;                                      //No crop left
    sps_parameter.frame_crop_right_offset = 0;									   //Crop right to get the correct frame width
    sps_parameter.frame_crop_top_offset = 0;                                       //No crop top
    sps_parameter.frame_crop_bottom_offset = 0;									   //Crop bottom to get the correct frame height
    sps_parameter.flags = sps_flags;
    sps_parameter.pOffsetForRefFrame = nullptr;                                    //Not send
    sps_parameter.pScalingLists = nullptr;                                         //Not send
    sps_parameter.pSequenceParameterSetVui = nullptr;                              //Not send
    
    //H.264 Specification Section 7.4.2.2:
    StdVideoH264PpsFlags pps_flags;
    pps_flags.transform_8x8_mode_flag = 0;                                         //Set to zero
    pps_flags.redundant_pic_cnt_present_flag = 0;                                  //Set to zero
    pps_flags.constrained_intra_pred_flag = 0;                                     //Set to zero
    pps_flags.deblocking_filter_control_present_flag = 0;                          //No deblocking
    pps_flags.weighted_bipred_idc_flag = 0;                                        //Set to zero
    pps_flags.weighted_pred_flag = 0;                                              //Set to zero
    pps_flags.pic_order_present_flag = 0;                                          //Set to zero
    pps_flags.entropy_coding_mode_flag = 0;                                        //Use CABAC for encoding
    pps_flags.pic_scaling_matrix_present_flag = 0;                                 //Dont send scaling matrix

    StdVideoH264PictureParameterSet pps_parameter;
    pps_parameter.seq_parameter_set_id = 0;                                        //Id of the sequence parameter set that is referenced
    pps_parameter.pic_parameter_set_id = 0;                                        //Id of the picture parameter set                                        
    pps_parameter.num_ref_idx_l0_default_active_minus1 = 0;                        //Set to zero
    pps_parameter.num_ref_idx_l1_default_active_minus1 = 0;                        //Set to zero
    pps_parameter.weighted_bipred_idc = STD_VIDEO_H264_WEIGHTED_BIPRED_IDC_DEFAULT; //Set to default
    pps_parameter.pic_init_qp_minus26 = 0;                                         //Set to zero
    pps_parameter.pic_init_qs_minus26 = 0;                                         //Set to zero
    pps_parameter.chroma_qp_index_offset = 0;                                      //Set to zero
    pps_parameter.second_chroma_qp_index_offset = 0;                               //Set to zero
    pps_parameter.flags = pps_flags;
    pps_parameter.pScalingLists = nullptr;                                         //Not send

    VkVideoEncodeH264SessionParametersAddInfoEXT add_info;
    add_info.sType = VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_ADD_INFO_EXT;
    add_info.pNext = nullptr;
    add_info.spsStdCount = 1;
    add_info.pSpsStd = &sps_parameter;
    add_info.ppsStdCount = 1;
    add_info.pPpsStd = &pps_parameter;

    VkVideoEncodeH264SessionParametersCreateInfoEXT encode_create_info;
    encode_create_info.sType = VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_CREATE_INFO_EXT;
    encode_create_info.pNext = nullptr;
    encode_create_info.maxSpsStdCount = 1;
    encode_create_info.maxPpsStdCount = 1;
    encode_create_info.pParametersAddInfo = &add_info;

    VkVideoSessionParametersCreateInfoKHR video_create_info;
    video_create_info.sType = VK_STRUCTURE_TYPE_VIDEO_SESSION_PARAMETERS_CREATE_INFO_KHR;
    video_create_info.pNext = &encode_create_info;
    video_create_info.videoSessionParametersTemplate = nullptr;
    video_create_info.videoSession = vkVideoSession;

    if(vkCreateVideoSessionParametersKHR(vkDevice, &video_create_info, nullptr, &vkVideoParameters) != VK_SUCCESS)
    {
        return false;
    }

    return true;
}

bool CreateImageMemory(VkImage image, VkDeviceMemory& memory)
{
	VkImageMemoryRequirementsInfo2 memoy_info;
	memoy_info.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_REQUIREMENTS_INFO_2;
	memoy_info.pNext = nullptr;
	memoy_info.image = image;

	VkMemoryRequirements2 memory_requirements;
	memory_requirements.sType = VK_STRUCTURE_TYPE_MEMORY_REQUIREMENTS_2;
	memory_requirements.pNext = nullptr;

	vkGetImageMemoryRequirements2(vkDevice, &memoy_info, &memory_requirements);

	uint32_t memoryTypeIndex = 0;

	if(!SelectMemory(memory_requirements.memoryRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, memoryTypeIndex))
	{
		return false;
	}

	VkMemoryAllocateInfo memoryAllocateInfo;
	memoryAllocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
	memoryAllocateInfo.pNext = nullptr;
	memoryAllocateInfo.allocationSize = memory_requirements.memoryRequirements.size;
	memoryAllocateInfo.memoryTypeIndex = memoryTypeIndex;

	if(vkAllocateMemory(vkDevice, &memoryAllocateInfo, nullptr, &memory) != VK_SUCCESS)
	{
		return false;
	}

	return true;
}

bool CreateImageView(VkImage image, VkFormat image_format, VkImageView& image_view)
{
	VkComponentMapping components;
	components.r = VK_COMPONENT_SWIZZLE_IDENTITY;
	components.g = VK_COMPONENT_SWIZZLE_IDENTITY;
	components.b = VK_COMPONENT_SWIZZLE_IDENTITY;
	components.a = VK_COMPONENT_SWIZZLE_IDENTITY;

	VkImageSubresourceRange subresource_range;
	subresource_range.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
	subresource_range.baseMipLevel = 0;
	subresource_range.levelCount = 1;
	subresource_range.baseArrayLayer = 0;
	subresource_range.layerCount = 1;

	VkImageViewCreateInfo create_info;
	create_info.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
	create_info.pNext = &vkVideoProfile;
	create_info.flags = 0;
	create_info.image = image;
	create_info.viewType = VK_IMAGE_VIEW_TYPE_2D;
	create_info.format = image_format;
	create_info.components = components;
	create_info.subresourceRange = subresource_range;

	if(vkCreateImageView(vkDevice, &create_info, nullptr, &image_view) != VK_SUCCESS)
	{
		return false;
	}

	return true;
}

bool CreateInputImage()
{
	VkExtent3D imageExtend;
	imageExtend.width = vkImageExtend.width;
	imageExtend.height = vkImageExtend.height;
	imageExtend.depth = 1;

	VkImageCreateInfo combined_create_info;
	combined_create_info.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
	combined_create_info.pNext = &vkVideoProfile;
	combined_create_info.flags = 0;
	combined_create_info.imageType = VK_IMAGE_TYPE_2D;
	combined_create_info.format = VK_FORMAT_G8_B8R8_2PLANE_444_UNORM_EXT;
	combined_create_info.extent = imageExtend;
	combined_create_info.mipLevels = 1;
	combined_create_info.arrayLayers = 1;
	combined_create_info.samples = VK_SAMPLE_COUNT_1_BIT;
	combined_create_info.tiling = VK_IMAGE_TILING_OPTIMAL;
	combined_create_info.usage = VK_IMAGE_USAGE_SAMPLED_BIT | VK_IMAGE_USAGE_VIDEO_ENCODE_SRC_BIT_KHR;
	combined_create_info.sharingMode = VK_SHARING_MODE_EXCLUSIVE;
	combined_create_info.queueFamilyIndexCount = 0;
	combined_create_info.pQueueFamilyIndices = nullptr;
	combined_create_info.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;

	if(vkCreateImage(vkDevice, &combined_create_info, nullptr, &vkInputImage) != VK_SUCCESS)
	{
		return false;
	}

	if(!CreateImageMemory(vkInputImage, vkInputMemory))
	{
		return false;
	}

	VkBindImageMemoryInfo bindInfo;
	bindInfo.sType = VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_INFO;
	bindInfo.pNext = nullptr;
	bindInfo.image = vkInputImage;
	bindInfo.memory = vkInputMemory;
	bindInfo.memoryOffset = 0;

	if(vkBindImageMemory2(vkDevice, 1, &bindInfo) != VK_SUCCESS)
	{
		return false;
	}

	if(!CreateImageView(vkInputImage, VK_FORMAT_G8_B8R8_2PLANE_444_UNORM_EXT, vkInputView))
	{
		return false;
	}

	return true;
}

bool CreateSlotImage()
{
	VkExtent3D imageExtend;
	imageExtend.width = vkImageExtend.width;
	imageExtend.height = vkImageExtend.height;
	imageExtend.depth = 1;

	VkImageCreateInfo combined_create_info;
	combined_create_info.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
	combined_create_info.pNext = &vkVideoProfile;
	combined_create_info.flags = 0;
	combined_create_info.imageType = VK_IMAGE_TYPE_2D;
	combined_create_info.format = VK_FORMAT_G8_B8R8_2PLANE_444_UNORM_EXT;
	combined_create_info.extent = imageExtend;
	combined_create_info.mipLevels = 1;
	combined_create_info.arrayLayers = 1;
	combined_create_info.samples = VK_SAMPLE_COUNT_1_BIT;
	combined_create_info.tiling = VK_IMAGE_TILING_OPTIMAL;
	combined_create_info.usage = VK_IMAGE_USAGE_SAMPLED_BIT | VK_IMAGE_USAGE_VIDEO_ENCODE_DPB_BIT_KHR;
	combined_create_info.sharingMode = VK_SHARING_MODE_EXCLUSIVE;
	combined_create_info.queueFamilyIndexCount = 0;
	combined_create_info.pQueueFamilyIndices = nullptr;
	combined_create_info.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;

	if(vkCreateImage(vkDevice, &combined_create_info, nullptr, &vkSlotImage) != VK_SUCCESS)
	{
		return false;
	}

	if(!CreateImageMemory(vkSlotImage, vkSlotMemory))
	{
		return false;
	}

	VkBindImageMemoryInfo bindInfo;
	bindInfo.sType = VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_INFO;
	bindInfo.pNext = nullptr;
	bindInfo.image = vkSlotImage;
	bindInfo.memory = vkSlotMemory;
	bindInfo.memoryOffset = 0;

	if(vkBindImageMemory2(vkDevice, 1, &bindInfo) != VK_SUCCESS)
	{
		return false;
	}

	if(!CreateImageView(vkSlotImage, VK_FORMAT_G8_B8R8_2PLANE_444_UNORM_EXT, vkSlotView))
	{
		return false;
	}

	return true;
}

bool CreateOutputBuffer()
{
	VkBufferCreateInfo create_info;
	create_info.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
	create_info.pNext = &vkVideoProfile;
	create_info.flags = 0;
	create_info.size = vkBufferSize;
	create_info.usage = VK_BUFFER_USAGE_VIDEO_ENCODE_DST_BIT_KHR;
	create_info.sharingMode = VK_SHARING_MODE_EXCLUSIVE;
	create_info.queueFamilyIndexCount = 0;
	create_info.pQueueFamilyIndices = nullptr;

	if(vkCreateBuffer(vkDevice, &create_info, nullptr, &vkOutputBuffer) != VK_SUCCESS)
	{
		return false;
	}

	VkBufferMemoryRequirementsInfo2 memory_info;
	memory_info.sType = VK_STRUCTURE_TYPE_BUFFER_MEMORY_REQUIREMENTS_INFO_2;
	memory_info.pNext = nullptr;
	memory_info.buffer = vkOutputBuffer;

	VkMemoryRequirements2 memory_requirements;
	memory_requirements.sType = VK_STRUCTURE_TYPE_MEMORY_REQUIREMENTS_2;
	memory_requirements.pNext = nullptr;

	vkGetBufferMemoryRequirements2(vkDevice, &memory_info, &memory_requirements);

	uint32_t memoryTypeIndex = 0;

	if(!SelectMemory(memory_requirements.memoryRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, memoryTypeIndex))
	{
		return false;
	}

	VkMemoryAllocateInfo memoryAllocateInfo;
	memoryAllocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
	memoryAllocateInfo.pNext = nullptr;
	memoryAllocateInfo.allocationSize = memory_requirements.memoryRequirements.size;
	memoryAllocateInfo.memoryTypeIndex = memoryTypeIndex;

	if(vkAllocateMemory(vkDevice, &memoryAllocateInfo, nullptr, &vkOutputMemory) != VK_SUCCESS)
	{
		return false;
	}

	VkBindBufferMemoryInfo bind_info;
	bind_info.sType = VK_STRUCTURE_TYPE_BIND_BUFFER_MEMORY_INFO;
	bind_info.pNext = nullptr;
	bind_info.buffer = vkOutputBuffer;
	bind_info.memory = vkOutputMemory;
	bind_info.memoryOffset = 0;

	if(vkBindBufferMemory2(vkDevice, 1, &bind_info) != VK_SUCCESS)
	{
		return false;
	}

	return true;
}

bool CreateCommandBuffer()
{
	VkCommandPoolCreateInfo createInfo;
	createInfo.sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
	createInfo.pNext = nullptr;
	createInfo.flags = 0;
	createInfo.queueFamilyIndex = vkQueueFamilyIndex;

	if(vkCreateCommandPool(vkDevice, &createInfo, nullptr, &vkCommandPool) != VK_SUCCESS)
	{
		return false;
	}

	VkCommandBufferAllocateInfo allocateInfo;
	allocateInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
	allocateInfo.pNext = nullptr;
	allocateInfo.commandPool = vkCommandPool;
	allocateInfo.level = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
	allocateInfo.commandBufferCount = 1;

	if(vkAllocateCommandBuffers(vkDevice, &allocateInfo, &vkCommandBuffer) != VK_SUCCESS)
	{
		return false;
	}

	return true;
}

bool CreateFence()
{
	VkFenceCreateInfo fenceCreateInfo;
	fenceCreateInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
	fenceCreateInfo.pNext = nullptr;
	fenceCreateInfo.flags = 0;

	if(vkCreateFence(vkDevice, &fenceCreateInfo, nullptr, &vkFence) != VK_SUCCESS)
	{
		return false;
	}

	return true;
}

//########################################################

void DestroyFence()
{
	vkDestroyFence(vkDevice, vkFence, nullptr);

	vkFence = nullptr;
}

void DestroyCommandBuffer()
{
	vkFreeCommandBuffers(vkDevice, vkCommandPool, 1, &vkCommandBuffer);
	vkDestroyCommandPool(vkDevice, vkCommandPool, nullptr);

	vkCommandBuffer = nullptr;
	vkCommandPool = nullptr;
}

void DestroyOutputBuffer()
{
	vkDestroyBuffer(vkDevice, vkOutputBuffer, nullptr);
	vkFreeMemory(vkDevice, vkOutputMemory, nullptr);

	vkOutputBuffer = nullptr;
	vkOutputMemory = nullptr;
}

void DestroySlotImage()
{
	vkDestroyImageView(vkDevice, vkSlotView, nullptr);
	vkDestroyImage(vkDevice, vkSlotImage, nullptr);
	vkFreeMemory(vkDevice, vkSlotMemory, nullptr);

	vkSlotView = nullptr;
	vkSlotImage = nullptr;
	vkSlotMemory = nullptr;
}

void DestroyInputImage()
{
	vkDestroyImageView(vkDevice, vkInputView, nullptr);
	vkDestroyImage(vkDevice, vkInputImage, nullptr);
	vkFreeMemory(vkDevice, vkInputMemory, nullptr);

	vkInputView = nullptr;
	vkInputImage = nullptr;
	vkInputMemory = nullptr;
}

void DestroyParameters()
{
	vkDestroyVideoSessionParametersKHR(vkDevice, vkVideoParameters, nullptr);

	vkVideoParameters = nullptr;

	for(VkDeviceMemory memory : vkVideoSessionMemory)
	{
		vkFreeMemory(vkDevice, memory, nullptr);
	}

	vkVideoSessionMemory.clear();
}

void DestroySession()
{
	vkDestroyVideoSessionKHR(vkDevice, vkVideoSession, nullptr);

	vkVideoSession = nullptr;
}

void DestroyDevice()
{
	vkDestroyDevice(vkDevice, nullptr);

	vkPhysicalDevice = nullptr;
	vkDevice = nullptr;

	vkQueueFamilyIndex = 0;
	vkQueue = nullptr;
}

void DestroyInstance()
{
	vkDestroyInstance(vkInstance, nullptr);

	vkInstance = nullptr;
}

//########################################################

void RecordEncodeCommands()
{
	const uint8_t sequence_parameter_id = 0;
	const uint8_t picture_parameter_id = 0;

	VkOffset2D encode_offset;
	encode_offset.x = 0;
	encode_offset.y = 0;

	VkExtent2D encode_extend;
	encode_extend.width = vkImageExtend.width;
	encode_extend.height = vkImageExtend.height;

	VkVideoPictureResourceKHR input_resource;
	input_resource.sType = VK_STRUCTURE_TYPE_VIDEO_PICTURE_RESOURCE_KHR;
	input_resource.pNext = nullptr;
	input_resource.codedOffset = encode_offset;
	input_resource.codedExtent = encode_extend;
	input_resource.baseArrayLayer = 0;
	input_resource.imageViewBinding = vkInputView;

	VkVideoPictureResourceKHR slot_resource;
	slot_resource.sType = VK_STRUCTURE_TYPE_VIDEO_PICTURE_RESOURCE_KHR;
	slot_resource.pNext = nullptr;
	slot_resource.codedOffset = encode_offset;
	slot_resource.codedExtent = encode_extend;
	slot_resource.baseArrayLayer = 0;
	slot_resource.imageViewBinding = vkSlotView;

	VkVideoReferenceSlotKHR slot_reference;
	slot_reference.sType = VK_STRUCTURE_TYPE_VIDEO_REFERENCE_SLOT_KHR;
	slot_reference.pNext = nullptr;
	slot_reference.slotIndex = 0;
	slot_reference.pPictureResource = &slot_resource;

	//--- Slot Info ----------------------------------------------------------------

	StdVideoEncodeH264PictureInfoFlags slot_flags;
	slot_flags.idr_flag = 0;
	slot_flags.is_reference_flag = 0;
	slot_flags.long_term_reference_flag = 0;

	StdVideoEncodeH264PictureInfo slot_parameters;
	slot_parameters.flags = slot_flags;
	slot_parameters.pictureType = STD_VIDEO_H264_PICTURE_TYPE_I;
	slot_parameters.frameNum = 0;
	slot_parameters.pictureOrderCount = 0;
	slot_parameters.long_term_pic_num = 0;
	slot_parameters.long_term_frame_idx = 0;

	VkVideoEncodeH264DpbSlotInfoEXT slot_info;
	slot_info.sType = VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_DPB_SLOT_INFO_EXT;
	slot_info.pNext = nullptr;
	slot_info.slotIndex = 0;
	slot_info.pStdPictureInfo = &slot_parameters;

	//--- Slice Info ----------------------------------------------------------------

	StdVideoEncodeH264RefMgmtFlags memory_flags;
	memory_flags.ref_pic_list_modification_l0_flag = 0;
	memory_flags.ref_pic_list_modification_l1_flag = 0;

	StdVideoEncodeH264RefMemMgmtCtrlOperations memory_operations;
	memory_operations.flags = memory_flags;
	memory_operations.refList0ModOpCount = 0;
	memory_operations.pRefList0ModOperations = nullptr;
	memory_operations.refList1ModOpCount = 0;
	memory_operations.pRefList1ModOperations = nullptr;
	memory_operations.refPicMarkingOpCount = 0;
	memory_operations.pRefPicMarkingOperations = nullptr;

	StdVideoEncodeH264SliceHeaderFlags slice_flags;
	slice_flags.idr_flag = 0;
	slice_flags.is_reference_flag = 0;
	slice_flags.num_ref_idx_active_override_flag = 0;
	slice_flags.no_output_of_prior_pics_flag = 1;
	slice_flags.long_term_reference_flag = 0;
	slice_flags.adaptive_ref_pic_marking_mode_flag = 0;
	slice_flags.no_prior_references_available_flag = 1;

	StdVideoEncodeH264SliceHeader slice_parameters;
	slice_parameters.flags = slice_flags;
	slice_parameters.slice_type = STD_VIDEO_H264_SLICE_TYPE_I;
	slice_parameters.seq_parameter_set_id = sequence_parameter_id;
	slice_parameters.pic_parameter_set_id = picture_parameter_id;
	slice_parameters.idr_pic_id = 0;
	slice_parameters.num_ref_idx_l0_active_minus1 = 0;
	slice_parameters.num_ref_idx_l1_active_minus1 = 0;
	slice_parameters.cabac_init_idc = STD_VIDEO_H264_CABAC_INIT_IDC_0;
	slice_parameters.disable_deblocking_filter_idc = STD_VIDEO_H264_DISABLE_DEBLOCKING_FILTER_IDC_DISABLED;
	slice_parameters.slice_alpha_c0_offset_div2 = 0;
	slice_parameters.slice_beta_offset_div2 = 0;
	slice_parameters.pMemMgmtCtrlOperations = &memory_operations;

	VkVideoEncodeH264NaluSliceEXT slice_info;
	slice_info.sType = VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_NALU_SLICE_EXT;
	slice_info.pNext = nullptr;
	slice_info.pSliceHeaderStd = &slice_parameters;
	slice_info.mbCount = vkBlockCount.width * vkBlockCount.height;
	slice_info.refFinalList0EntryCount = 1;
	slice_info.pRefFinalList0Entries = &slot_info;
	slice_info.refFinalList1EntryCount = 1;
	slice_info.pRefFinalList1Entries = &slot_info;

	//------------------------------------------------------------------------------

	VkVideoEncodeH264VclFrameInfoEXT frame_info;
	frame_info.sType = VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_VCL_FRAME_INFO_EXT;
	frame_info.pNext = nullptr;
	frame_info.refDefaultFinalList0EntryCount = 1;
	frame_info.pRefDefaultFinalList0Entries = &slot_info;
	frame_info.refDefaultFinalList1EntryCount = 1;
	frame_info.pRefDefaultFinalList1Entries = &slot_info;
	frame_info.naluSliceEntryCount = 1;
	frame_info.pNaluSliceEntries = &slice_info;
	frame_info.pCurrentPictureInfo = &slot_info;

	VkVideoEncodeInfoKHR encode_info;
	encode_info.sType = VK_STRUCTURE_TYPE_VIDEO_ENCODE_INFO_KHR;
	encode_info.pNext = &frame_info;
	encode_info.flags = VK_VIDEO_ENCODE_RESERVED_0_BIT_KHR;
	encode_info.qualityLevel = 0;
	encode_info.codedExtent = encode_extend;
	encode_info.dstBitstreamBuffer = vkOutputBuffer;
	encode_info.dstBitstreamBufferOffset = 0;
	encode_info.dstBitstreamBufferMaxRange = vkBufferSize;
	encode_info.srcPictureResource = input_resource;
	encode_info.pSetupReferenceSlot = &slot_reference;
	encode_info.referenceSlotCount = 1;
	encode_info.pReferenceSlots = &slot_reference;
	encode_info.precedingExternallyEncodedBytes = 0;

	vkCmdEncodeVideoKHR(vkCommandBuffer, &encode_info);
}

void RecordSetupCommands()
{
	VkImageSubresourceRange image_range;
	image_range.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
	image_range.baseMipLevel = 0;
	image_range.levelCount = 1;
	image_range.baseArrayLayer = 0;
	image_range.layerCount = 1;

	VkImageMemoryBarrier input_barrier;
	input_barrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
	input_barrier.pNext = nullptr;
	input_barrier.srcAccessMask = 0;
	input_barrier.dstAccessMask = 0;
	input_barrier.oldLayout = VK_IMAGE_LAYOUT_UNDEFINED;
	input_barrier.newLayout = VK_IMAGE_LAYOUT_VIDEO_ENCODE_SRC_KHR;
	input_barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
	input_barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
	input_barrier.image = vkInputImage;
	input_barrier.subresourceRange = image_range;

	VkImageMemoryBarrier slot_barrier;
	slot_barrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
	slot_barrier.pNext = nullptr;
	slot_barrier.srcAccessMask = 0;
	slot_barrier.dstAccessMask = 0;
	slot_barrier.oldLayout = VK_IMAGE_LAYOUT_UNDEFINED;
	slot_barrier.newLayout = VK_IMAGE_LAYOUT_VIDEO_ENCODE_DPB_KHR;
	slot_barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
	slot_barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
	slot_barrier.image = vkSlotImage;
	slot_barrier.subresourceRange = image_range;

	std::array<VkImageMemoryBarrier, 2> image_barriers;
	image_barriers[0] = input_barrier;
	image_barriers[1] = slot_barrier;

	vkCmdPipelineBarrier(vkCommandBuffer, VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT, 0, 0, nullptr, 0, nullptr, (uint32_t)image_barriers.size(), image_barriers.data());

	VkOffset2D encode_offset;
	encode_offset.x = 0;
	encode_offset.y = 0;

	VkExtent2D encode_extend;
	encode_extend.width = vkImageExtend.width;
	encode_extend.height = vkImageExtend.height;

	VkVideoPictureResourceKHR slot_resource;
	slot_resource.sType = VK_STRUCTURE_TYPE_VIDEO_PICTURE_RESOURCE_KHR;
	slot_resource.pNext = nullptr;
	slot_resource.codedOffset = encode_offset;
	slot_resource.codedExtent = encode_extend;
	slot_resource.baseArrayLayer = 0;
	slot_resource.imageViewBinding = vkSlotView;

	VkVideoReferenceSlotKHR slot_reference;
	slot_reference.sType = VK_STRUCTURE_TYPE_VIDEO_REFERENCE_SLOT_KHR;
	slot_reference.pNext = nullptr;
	slot_reference.slotIndex = 0;
	slot_reference.pPictureResource = &slot_resource;

	VkVideoBeginCodingInfoKHR begin_info;
	begin_info.sType = VK_STRUCTURE_TYPE_VIDEO_BEGIN_CODING_INFO_KHR;
	begin_info.pNext = nullptr;
	begin_info.flags = 0;
	begin_info.codecQualityPreset = VK_VIDEO_CODING_QUALITY_PRESET_NORMAL_BIT_KHR;
	begin_info.videoSession = vkVideoSession;
	begin_info.videoSessionParameters = vkVideoParameters;
	begin_info.referenceSlotCount = 1;
	begin_info.pReferenceSlots = &slot_reference;

	vkCmdBeginVideoCodingKHR(vkCommandBuffer, &begin_info);

	RecordEncodeCommands();

	VkVideoEndCodingInfoKHR end_info;
	end_info.sType = VK_STRUCTURE_TYPE_VIDEO_END_CODING_INFO_KHR;
	end_info.pNext = nullptr;
	end_info.flags = 0;

	vkCmdEndVideoCodingKHR(vkCommandBuffer, &end_info);
}

bool RecordCommands()
{
	VkCommandBufferBeginInfo commandBeginInfo;
	commandBeginInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
	commandBeginInfo.pNext = nullptr;
	commandBeginInfo.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;
	commandBeginInfo.pInheritanceInfo = nullptr;

	if(vkBeginCommandBuffer(vkCommandBuffer, &commandBeginInfo) != VK_SUCCESS)
	{
		return false;
	}

	RecordSetupCommands();

	if(vkEndCommandBuffer(vkCommandBuffer) != VK_SUCCESS)
	{
		return false;
	}

	return true;
}

bool SubmitCommandBuffer()
{
	VkSubmitInfo submit_info;
	submit_info.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
	submit_info.pNext = nullptr;
	submit_info.waitSemaphoreCount = 0;
	submit_info.pWaitSemaphores = nullptr;
	submit_info.pWaitDstStageMask = nullptr;
	submit_info.commandBufferCount = 1;
	submit_info.pCommandBuffers = &vkCommandBuffer;
	submit_info.signalSemaphoreCount = 0;
	submit_info.pSignalSemaphores = nullptr;

	//This is the place where the crash happens.
	if(vkQueueSubmit(vkQueue, 1, &submit_info, vkFence) != VK_SUCCESS)
	{
		return false;
	}

	if(vkWaitForFences(vkDevice, 1, &vkFence, VK_TRUE, 0xFFFFFFFF) != VK_SUCCESS)
	{
		return false;
	}

	return true;
}

//########################################################

void main()
{
	if(volkInitialize() != VK_SUCCESS)
	{
		return;
	}

	if(!CreateInstance())
	{
		return;
	}

	volkLoadInstance(vkInstance);

	if(!CreateDevice())
	{
		return;
	}

	if(!CheckProfiles())
	{
		return;
	}

	ComputeDimensions();

	if(!CreateSession())
	{
		return;
	}

	if(!BindSessionMemory())
	{
		return;
	}

	if(!CreateParameters())
	{
		return;
	}

	if(!CreateInputImage())
	{
		return;
	}

	if(!CreateSlotImage())
	{
		return;
	}

	if(!CreateOutputBuffer())
	{
		return;
	}

	if(!CreateCommandBuffer())
	{
		return;
	}

	if(!CreateFence())
	{
		return;
	}

	if(!RecordCommands())
	{
		return;
	}

	if(!SubmitCommandBuffer())
	{
		return;
	}

	DestroyFence();
	DestroyCommandBuffer();
	DestroyOutputBuffer();
	DestroySlotImage();
	DestroyInputImage();
	DestroyParameters();
	DestroySession();
	DestroyDevice();
	DestroyInstance();
}